/******************************************************************************/
/*                                                                            */
/* Version 1.16                                                               */
/* Author: Arjan van Vught                 Date: October   2001               */
/*                                                                            */
/******************************************************************************/
/* Version 1.0   First release NT                                             */
/* Version 1.1   Own written SMTP client functions                            */
/* Version 1.2   Code clean up - removed additional debug info                */
/* Version 1.3   Updated for MQSeries Integrator 2.0.1                        */
/* Version 1.4   First release AIX                                            */
/* Version 1.5   Sendmail support                                             */
/* Version 1.6   Solaris Ready (not tested)                                   */
/* Version 1.7   New properties added                                         */
/* Version 1.8   Bug with CCI_BUFFER_TOO_SMALL solved                         */
/* Version 1.9   sprintf for Date field changed                               */
/* Version 1.10  New property SMTP server port number added                   */
/* Version 1.11  Updated for MQSeries Integrator 2.0.2                        */
/* Version 1.12  Changed the "to" and "from" fields in the input xml so       */
/*               they aren't reserved words                                   */
/* Version 1.13  Enhanced the Body element with Line elements,                */
/*               every Line element is a new line in the mail body            */
/* Version 1.14  First release Sun Solaris                                    */
/* Version 1.15  First release Linux                                          */
/* Version 1.16  First release HP-UX Ready                                    */
/******************************************************************************/
/*                                                                            */
/* Note: This plugin node implementation is not necessarily portable          */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/
/* Includes */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>                                                             /* 15/05/2001 */
#include <wchar.h>

#ifdef __LCLINT__																					/* 02/08/2001 */
#ifdef SENDMAIL																					/* 02/08/2001 */
#undef SENDMAIL																					/* 02/08/2001 */
#define SMTP																						/* 02/08/2001 */
#endif																								/* 02/08/2001 */
#endif																								/* 02/08/2001 */

#include "SendMailPlugIn.h"

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for attribute and link to the context                     */
/*                                                                            */
/******************************************************************************/
ATTRIBUTE_TBL_ENTRY *insertAttributeTableEntry(NODE_CONTEXT_ST *context, CciChar *nodeName, CciChar *attrName, int type)
{
   /*@unused@*/static char *functionName = (char *)"insertAttributeTableEntry()";			/* 28/07/2001 */

   ATTRIBUTE_TBL_ENTRY *entry = NULL;
   int attrSize = 4;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   /* Allocate storage for the attribute entry */
   entry = (ATTRIBUTE_TBL_ENTRY *)malloc(sizeof(ATTRIBUTE_TBL_ENTRY));			/* 19/07/2001 */
   if (entry) {
      entry->next = 0;
      entry->type = type;

      switch (entry->type) {
         case CNI_TYPE_INTEGER:
            attrSize = 4;
            break;
         case CNI_TYPE_STRING:
            attrSize = MAX_STRING_SIZE;
				break;																				/* 28/07/2001 */
         default:
            /* Error: Invalid datatype */
            break;
      } /* switch (entry->type) */

      /* Acquire and clear storage for attribute value */
      entry->valptr = malloc(attrSize);

      if (entry->valptr) memset(entry->valptr, 0, attrSize);

      /* Store the attribute name */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      CciCharCpy(entry->name, attrName);
#else
      wcscpy(entry->name, attrName);
#endif

      /* Link an existing previous element to this one */
      if (context->attrListPrevious) context->attrListPrevious->next = entry;
      else if ((context->attrListHead) == 0) context->attrListHead = entry;

      /* Save the pointer to the previous element */
      context->attrListPrevious = entry;
   }
   else {
         /* Error: Unable to allocate memory */
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, entry = %p\n", functionName, entry);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   return(entry);

}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an input terminal entry and link it to the context    */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insertInputTerminalListEntry(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   /*@unused@*/static char *functionName = (char *)"insertInputTerminalListEntry()";		/* 28/07/2001 */	

   TERMINAL_LIST_ENTRY *entry = NULL;
   int rc;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   entry = malloc(sizeof(TERMINAL_LIST_ENTRY));

   if (entry) {
      /* This entry is the current end of the list */
      entry->next = 0;

    /* Store the terminal name */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      CciCharCpy(entry->name, terminalName);
#else
      wcscpy(entry->name, terminalName);
#endif

      /* Create terminal and save its handle */
      entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, terminalName);

      /* Link an existing previous element to this one */
      if (context->inputTerminalListPrevious)
	      context->inputTerminalListPrevious->next = entry;
      else if ((context->inputTerminalListHead) == 0)
	      context->inputTerminalListHead = entry;

      /* Save the pointer to the previous element */
      context->inputTerminalListPrevious = entry;

   } /* if (entry) */

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, entry = %p\n", functionName, entry);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

  return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an output terminal and link it to the context         */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insertOutputTerminalListEntry(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   /*@unused@*/static char *functionName = (char *)"insertOutputTerminalListEntry()";/* 28/07/2001 */

   TERMINAL_LIST_ENTRY *entry = NULL;
   int rc;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);	/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   entry = malloc(sizeof(TERMINAL_LIST_ENTRY));

   if (entry) {
      /* This entry is the current end of the list */
      entry->next = 0;

      /* Store the terminal name */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      CciCharCpy(entry->name, terminalName);
#else
      wcscpy(entry->name, terminalName);
#endif

      /* Create terminal and save its handle */
      entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, terminalName);

      /* Link an existing previous element to this one */
      if (context->outputTerminalListPrevious)
	      context->outputTerminalListPrevious->next = entry;
      else if ((context->outputTerminalListHead) == 0)
	      context->outputTerminalListHead = entry;

      /* Save the pointer to the previous element */
      context->outputTerminalListPrevious = entry;
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, entry = %p\n", functionName, entry);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named input terminal                     */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   static char *functionName = (char *)"getInputTerminalHandle()";				/* 28/07/2001 */

   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->inputTerminalListHead;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

  /* Scan the list for a matching terminal name */
  while (p != 0) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(p->name, terminalName) == 0) {
#else
      if (wcscmp(p->name, terminalName) == 0) {
#endif
         /* Match found; return it */
#ifdef _DEBUG
         fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, Match found\n", functionName);/* 28/07/2001 */
         fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
         return(p->handle);
      }
      p = p->next;
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, No match was found, rc = NULL\n", functionName);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   /* No match was found */
   return(NULL);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named output terminal                    */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   static char *functionName = (char *)"getOutputTerminalHandle()";				/* 28/07/2001 */

   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->outputTerminalListHead;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   /* Scan the list for a matching terminal name */
   while (p != 0) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(p->name, terminalName) == 0) {
#else
      if (wcscmp(p->name, terminalName) == 0) {
#endif
         /* Match found; return it */
#ifdef _DEBUG
         fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, Match found\n", functionName);/* 28/07/2001 */
         fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
         return(p->handle);
      }
      p = p->next;

   } /* while (p != 0) */

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, No match was found, rc = 0\n", functionName);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   /* No match was found */
   return(NULL);
}

/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttriubteName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniCreateNodeContext() */
/*                                                                            */
/******************************************************************************/
CciContext *createNodeContext(CciFactory *factoryObject, CciChar *nodeName, CciNode *nodeObject)
{
   /*@unused@*/static char *functionName = (char *)"createNodeContext()";						/* 28/07/2001 */
   NODE_CONTEXT_ST *p = NULL;

   /* Allocate a pointer to the local context */
   p = (NODE_CONTEXT_ST *)malloc(sizeof(NODE_CONTEXT_ST));

   /* If malloc was successful */
   if (p) {
      /* Clear the context area */
      memset(p, 0, sizeof(NODE_CONTEXT_ST)); /* Clear the context area */
      /* SMTP support                                                                       */
      p->SMTPHostName          = NULL;
      p->SMTPHostPort          = (unsigned short)25;
      p->SMTPFrom              = NULL;
      p->SMTPFullFrom          = NULL;
      p->SMTPMailFrom          = NULL;
      /* sendmail support                                                                   */
      p->SendmailOptionalParms = NULL;
      /* Basic                                                                              */
      p->BasicProperties       = FALSE;
      p->BasicBcc              = NULL;
      p->BasicPriority         = (char)3;  /* 3 = Normal */                   /* 13/04/2001 */

#ifdef _DEBUG
#if defined(__CCI_LINUX__)																		/* 16/09/2001 */
      p->debugFile = fopen("/tmp/SendMailPlugInDebug.txt", "w");
      fprintf(p->debugFile, "SendMailPlugIn for MQSeries Integrator V2.0.2 for Linux\n");
#elif defined(__CCI_SOLARIS__)																	
      p->debugFile = fopen("/tmp/SendMailPlugInDebug.txt", "w");
      fprintf(p->debugFile, "SendMailPlugIn for MQSeries Integrator V2 for Solaris\n");
#elif defined(__CCI_HPUX__)																	/* 19/07/2001 */
      p->debugFile = fopen("/tmp/SendMailPlugInDebug.txt", "w");
      fprintf(p->debugFile, "SendMailPlugIn for MQSeries Integrator V2.0.2 for HP-UX\n"); 
#elif defined(__CCI_AIX__)
      p->debugFile = fopen("/tmp/SendMailPlugInDebug.txt", "w");
      fprintf(p->debugFile, "SendMailPlugIn for MQSeries Integrator V2 for AIX\n"); 
#else
      p->debugFile = fopen("c:\\SendMailPlugInDebug.txt", "w");
      fprintf(p->debugFile, "SendMailPlugIn for MQSeries Integrator V2 for Windows NT\n");
#endif
      fprintf(p->debugFile, "-> %s\n", functionName);									/* 28/07/2001 */
      fflush(p->debugFile);
#endif

      /* Save our node object pointer in our context */
	   p->nodeObject = nodeObject;

    /* Save our node name */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      CciCharNCpy((CciChar*)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);
#else
      wcsncpy((CciChar *)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);
#endif

	   /* Verify the name of the node is supported by the factory */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(nodeName, constPlugInNode) == 0) {
#else
	   if (wcscmp(nodeName, L"SendMailPlugInNode") == 0) {
#endif
         /* Create attributes */
         /* General */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         {
            const CciChar *ucsAttr = CciString("Bcc") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("Priority") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("SetPriority") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
#else
         insertAttributeTableEntry(p, nodeName, L"Bcc",         CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"Priority",    CNI_TYPE_STRING);/* 13/04/2001 */
         insertAttributeTableEntry(p, nodeName, L"SetPriority", CNI_TYPE_STRING);/* 13/04/2001 */
         /* SMTP support */
#endif
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         {
            const CciChar *ucsAttr = CciString("ServerName") ;
            insertAttributeTableEntry(p, nodeName, (CciChar *)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("ServerPort") ;
            insertAttributeTableEntry(p, nodeName, (CciChar *)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("SMTPFrom") ;
            insertAttributeTableEntry(p, nodeName, (CciChar *)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("SMTPFullFrom") ;
            insertAttributeTableEntry(p, nodeName, (CciChar *)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
#else
         insertAttributeTableEntry(p, nodeName, L"ServerName",      CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"ServerPort",      CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"SMTPFrom",        CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"SMTPFullFrom",    CNI_TYPE_STRING);
#endif
         /* sendmail support */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         {
            const CciChar *ucsAttr = CciString("C") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("F") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("f") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("h") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
         {
            const CciChar *ucsAttr = CciString("X") ;
            insertAttributeTableEntry(p, nodeName, (CciChar*)ucsAttr, CNI_TYPE_STRING);
            free((void *)ucsAttr) ;
         }
#else
         insertAttributeTableEntry(p, nodeName, L"C", CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"F", CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"f", CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"h", CNI_TYPE_STRING);
         insertAttributeTableEntry(p, nodeName, L"X", CNI_TYPE_STRING);
#endif
         /* Create terminals */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         {
            const CciChar* ucsIn = CciString("in") ;
            insertInputTerminalListEntry(p, (CciChar *)ucsIn);
            free((void *)ucsIn) ;
         }
    
         {
            const CciChar* ucsOut = CciString("out") ;
            insertOutputTerminalListEntry(p, (CciChar *)ucsOut);
            free((void *)ucsOut) ;
         }

         {
            const CciChar* ucsFailure = CciString("failure") ;
            insertOutputTerminalListEntry(p, (CciChar *)ucsFailure);
            free((void *)ucsFailure) ;
         }
#else
	      insertInputTerminalListEntry(p,  L"in");
	      insertOutputTerminalListEntry(p, L"out");
	      insertOutputTerminalListEntry(p, L"failure");
#endif
	   }
	   else{
	      /* Error handling if name of node is not supported by factory */
	   }
   }
   else {
      /* Error handling if malloc failed */
   }

#ifdef _DEBUG
    fprintf(p->debugFile, "<- %s\n", functionName);									/* 28/07/2001 */
    fflush(p->debugFile);
#endif

   /* Return the pointer to the local context */
   return(p);
}


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniDeleteNodeContext() */
/*                                                                            */
/* This function is called when an instance of a node is deleted.             */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext* context)
{
  /*@unused@*/static char *functionName = (char *)"deleteNodeContext()";						/* 28/07/2001 */

  NODE_CONTEXT_ST *p = (NODE_CONTEXT_ST *)context;

#ifdef _DEBUG
  fprintf(p->debugFile, "<-> %s\n", functionName);										/* 28/07/2001 */
  fclose(p->debugFile);
#endif

  free(p->SendmailOptionalParms);
  free(p->SMTPHostName);

  free((NODE_CONTEXT_ST *)context);

  return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int getAttribute(CciContext *context, CciChar *attrName, CciChar *buffer, int bufsize)
{
   static char *functionName = (char *)"getAttribute()";								/* 28/07/2001 */

   ATTRIBUTE_TBL_ENTRY *p = ((NODE_CONTEXT_ST *)context)->attrListHead;

   wchar_t wstring[256];
   char    string[256];  

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);	/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   while (p != 0) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(p->name, attrName) == 0) {
#else
      if (wcscmp(p->name, attrName) == 0) {
#endif
         switch(p->type) {
            case CNI_TYPE_INTEGER:
               sprintf((char*)&string, "%d", *((int *)p->valptr));
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
               {
                  const CciChar* ucsAttrValue = CciString(string);
                  CciCharNCpy(buffer, ucsAttrValue, min(bufsize, 256));
                  free((void *)ucsAttrValue);
               }
#else
               mbstowcs((wchar_t*)&wstring, (char *)&string, 256); 
               wcsncpy(buffer, (wchar_t *)&wstring, (size_t)min(bufsize, 256));
#endif
               break;
            case CNI_TYPE_STRING:
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
               {
                  const CciChar* ucsAttrValue = CciString((char *)p->valptr);
                  CciCharNCpy(buffer, ucsAttrValue, min(MAX_STRING_SIZE, bufsize));
                  free((void *)ucsAttrValue);
               }
#else
               memcpy(buffer, p->valptr, (size_t)min(MAX_STRING_SIZE, bufsize));
#endif
            default:
            /* Error: Invalid datatype */
            break;
         }

#ifdef _DEBUG
         fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 0\n", functionName);/* 28/07/2001 */
         fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
         /* Normal return; attribute value returned */ 
         return(0);
      }

      p = p->next;
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 0\n", functionName);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
   /* Attribute not owned by plugin */
   return(1); 
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniSetAttribute() */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *context, CciChar *attrName, CciChar *attrValue)
{
   static char *functionName = (char *)"setAttribute()";								/* 28/07/2001 */

   ATTRIBUTE_TBL_ENTRY *p = ((NODE_CONTEXT_ST *)context)->attrListHead;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
   
   while (p != 0) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(p->name, attrName) == 0) {
#else
      if (wcscmp(p->name, attrName) == 0) {
#endif
         switch(p->type) {
            case CNI_TYPE_INTEGER:
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
               {
                  const char *mbAttrValue = mbString(attrValue);
                  *((int *)p->valptr) = atoi(mbAttrValue);
                  free((void *)mbAttrValue);
               }
#elif defined(__CCI_AIX__)
               *((int *)p->valptr) = watoi(attrValue);
#else
               *((int *)p->valptr) = _wtoi(attrValue);
#endif
               break;
            case CNI_TYPE_STRING:
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
               {
                  const char *mbAttrValue = mbString(attrValue);
                  memcpy(p->valptr, mbAttrValue, MAX_STRING_SIZE);
                  free((void *)mbAttrValue);
               }
#else
               memcpy(p->valptr, attrValue, MAX_STRING_SIZE); 
#endif
					break;																			/* 28/07/2001 */
            default:
               /* Error: Invalid datatype */
               break;
         }
#ifdef _DEBUG
         fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 0\n", functionName);/* 28/07/2001 */
         fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
         /* Normal return; attribute value stored by plugin */
         return(0);
      }

      p = p->next;
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 1\n", functionName);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
   /* Attribute not owned by plugin */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int getAttributeName(CciContext* context, int index, CciChar* buffer, int bufferSize)
{
   static char *functionName = (char *)"getAttributeName()";						/* 28/07/2001 */

   ATTRIBUTE_TBL_ENTRY *p = ((NODE_CONTEXT_ST *)context)->attrListHead;
   int count = 0;

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "-> %s\n", functionName);
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
    
   while (p != 0) {
      if (count == index) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         CciCharCpy(buffer, p->name);
#else
         wcscpy(buffer, p->name);
#endif

#ifdef _DEBUG
         fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 0\n", functionName);/* 28/07/2001 */
         fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif
         /* Normal return; attribute name returned */
         return(0);
      }
      count++;
      p = p->next;
   }

#ifdef _DEBUG
   fprintf(((NODE_CONTEXT_ST *)context)->debugFile, "<- %s, rc = 1\n", functionName);/* 28/07/2001 */
   fflush(((NODE_CONTEXT_ST *)context)->debugFile);
#endif

   /* Attribute not owned by the plugin node; defer to base */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                         cniEvaluate() */
/*                                                                            */
/******************************************************************************/
void evaluate(CciContext *context, CciMessage *destinationList, CciMessage *exceptionList, CciMessage *message)
{
   /*@unused@*/static char *functionName = (char *)"evaluate()";					/* 28/07/2001 */
  
   CCI_EXCEPTION_ST exception_st;
   NODE_CONTEXT_ST *p = (NODE_CONTEXT_ST *)context;
    
   NodeStringBuffer *To    = NULL;
   NodeStringBuffer *Cc    = NULL;
   NodeStringBuffer *Bcc   = NULL;
	NodeStringBuffer *Line  = NULL;

#ifdef SENDMAIL
   char *aConfigFileName = NULL;
   char *aFullName       = NULL;
   char *aNameFrom       = NULL;
   char *aHopCount       = NULL;
   char *aLogFileName    = NULL;
#endif

   char *From        = NULL;
   char *Subject     = NULL;
   char *Body        = NULL;
   char *pTmp        = NULL;
   char *aValueChar  = NULL;
   char *ErrorString = NULL;
   char *Parm0       = NULL;

   unsigned char  Checksum = 0;                                               /* 19/10/2001 */

   static char MailDate[128];
   static char MailMsgId[64];
   static char pStr[8];
#ifdef SENDMAIL
   static char OptionalParms[10240];
#endif
 
   CciChar  AttributeBuffer[256];
   CciChar  elementName[256];
   CciChar *valueChar    = NULL;                                              /* 21/04/2001 */
   CciChar *pErrorString = NULL;
     
   CciTerminal *outputTerminalObject = NULL;

   CciMessage *outMsg = NULL;

   CciElement *rootElement    = NULL;
   CciElement *mqmdFolder     = NULL;
   CciElement *Element        = NULL;
	CciElement *MailBodyLine   = NULL;														/* 20/06/2001 */
   CciElement *newElement     = NULL;
   CciElement *outRootElement = NULL;
   CciElement *bodyChild      = NULL;
   CciElement *lastChild      = NULL;

   CciSize elementNameSize;

   struct CciDate       valueDate;
   struct CciTime       valueTime;
   struct CciByteArray *Blob = NULL;
   
   int rc            = -1;
   int NoOfBytes     =  0;
   int i             =  0;
   int retvalue      =  0;
   int rcTcp         =  0;
   int valueCharSize =  1024;

   unsigned short Parm1 = 0;

   struct tm *newtime = NULL;                                                 /* 15/05/2001 */
   time_t ltime = 0;                                                          /* 15/05/2001 */
   char MQMDvalid = FALSE;                                                    /* 15/05/2001 */

#ifdef _DEBUG
   fprintf(p->debugFile, "-> %s\n", functionName);
   fflush(p->debugFile);
#ifdef WIN32
 	DisplayMemoryStatus();
#endif
#endif

   if (!p->BasicProperties) {
      p->BasicProperties = TRUE;
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("Bcc") ;									/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"Bcc", AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
#endif																								/* 22/07/2001 */
		if (rc == 0)  {																			/* 22/07/2001 */
         p->BasicBcc = UnicodeToASCII(AttributeBuffer);
#ifdef _DEBUG
         fprintf(p->debugFile, "BasicBcc = %s\n", p->BasicBcc);
         fflush(p->debugFile);
#endif
      } /* if (rc == 0)  */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("SetPriority");							/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"SetPriority", AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
#ifdef _DEBUG
         fprintf(p->debugFile, "SetPriority = %S\n", AttributeBuffer);
         fflush(p->debugFile);
#endif
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         if (CciCharCmp(AttributeBuffer, (CciChar *)constYes) == 0) {			/* 22/07/2001 */
#else
         if (wcscmp(AttributeBuffer, L"yes") == 0) {
#endif
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
				{																						/* 22/07/2001 */
					const CciChar *ucsAttr = CciString("Priority");						/* 22/07/2001 */
					rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
					free((void *)ucsAttr) ;														/* 22/07/2001 */
				}																						/* 22/07/2001 */
#else																									/* 22/07/2001 */
            rc = getAttribute(context, L"Priority", AttributeBuffer, sizeof(AttributeBuffer));
#endif
				if (rc == 0)  {																	/* 22/07/2001 */
               p->BasicPriority = GetPriorityNumber(AttributeBuffer);
#ifdef _DEBUG
               fprintf(p->debugFile, "Priority      = [%S]\n", AttributeBuffer);
               fprintf(p->debugFile, "BasicPriority = %d\n", p->BasicPriority);
               fflush(p->debugFile);
#endif
            } /* if (rc == 0)  */
         } /* if (wcscmp(AttributeBuffer, L"yes") == 0) */
      } /* if (rc == 0) */


   } /* if (!p->BasicProperties) */

#ifdef SMTP
   if (p->SMTPHostName == NULL) {
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("ServerName");							/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"ServerName", AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
#endif																								/* 22/07/2001 */
		if (rc == 0)  {																			/* 22/07/2001 */
         p->SMTPHostName = UnicodeToASCII(AttributeBuffer);
      } /* if (rc == 0) */

#ifdef _DEBUG
      fprintf(p->debugFile, "Using SMTP with hostname = [%s]\n", p->SMTPHostName);
      fflush(p->debugFile);
#endif

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("ServerPort");							/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"ServerPort", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         pTmp = UnicodeToASCII(AttributeBuffer);
         p->SMTPHostPort = (unsigned short)atoi(pTmp);
         free(pTmp);
      } /* if (rc == 0) */

#ifdef _DEBUG
      fprintf(p->debugFile, "         AttributeBuffer = [%S]\n", AttributeBuffer);
      fprintf(p->debugFile, "             Port Number = [%d]\n", p->SMTPHostPort);
      fflush(p->debugFile);
#endif

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("SMTPFrom")	;							/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"SMTPFrom", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         p->SMTPFrom = UnicodeToASCII(AttributeBuffer);
#ifdef _DEBUG
         fprintf(p->debugFile, "SMTPFrom = [%s]\n", p->SMTPFrom);
         fflush(p->debugFile);
#endif
      } /* if (rc == 0)  */
      else
         p->SMTPFrom = NULL;
   
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("SMTPFullFrom");						/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
		rc = getAttribute(context, L"SMTPFullFrom", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
          p->SMTPFullFrom = UnicodeToASCII(AttributeBuffer);
#ifdef _DEBUG
         fprintf(p->debugFile, "SMTPFullFrom = [%s]\n", p->SMTPFullFrom);
         fflush(p->debugFile);
#endif
      } /* if (rc == 0)  */
      else
         p->SMTPFullFrom = NULL;

      if ( (p->SMTPFrom != NULL) && (strlen(p->SMTPFrom) != 0) && (p->SMTPFrom[0] != ' ') ) {
         if ( (p->SMTPFullFrom != NULL) && (strlen(p->SMTPFullFrom) != 0) && (p->SMTPFullFrom[0] != ' ') ) {
            p->SMTPMailFrom = (char *)malloc(sizeof(char) * (strlen(p->SMTPFrom) + strlen(p->SMTPFullFrom) + 1 + strlen(" <> ")));
       
				strcpy(p->SMTPMailFrom, p->SMTPFullFrom);									/* 25/09/2001 */
            strcat(p->SMTPMailFrom, " <");
	         strcat(p->SMTPMailFrom, p->SMTPFrom);										/* 25/09/2001 */
            strcat(p->SMTPMailFrom, ">");

#ifdef _DEBUG
            fprintf(p->debugFile, "p->SMTPMailFrom = [%s]\n", p->SMTPMailFrom);
            fflush(p->debugFile);
#endif
         } /* if ( (p->SMTPFullFrom != NULL) && (strlen(p->SMTPFullFrom) != 0) && (p->SMTPFullFrom[0] != ' ') ) */
         else
            p->SMTPMailFrom = p->SMTPFrom;
      } /* if ( (p->SMTPFrom != NULL) && (strlen(p->SMTPFrom) != 0) && (p->SMTPFrom[0] != ' ') ) */

   } /* if (p->SMTPHostName == NULL) */
#endif /* #ifdef SMTP */

#ifdef SENDMAIL
   if (p->SendmailOptionalParms == NULL) {

      memset(OptionalParms, 0, sizeof(OptionalParms));

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("C");										/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"C", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         aConfigFileName = UnicodeToASCII(AttributeBuffer);
         if (aConfigFileName[0] != '\0') {
            strcat(OptionalParms, " -C ");
            strcat(OptionalParms, aConfigFileName);
         }
      } /* if (rc == 0) */
      
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("F");										/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"F", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         aFullName = UnicodeToASCII(AttributeBuffer);
         if (aFullName[0] != '\0') {
            if ( (aFullName[0] == '\'') && (aFullName[strlen(aFullName) - 1] == '\'') ) {
               strcat(OptionalParms, " -F ");
               strcat(OptionalParms, aFullName);
            }
            else {
               strcat(OptionalParms, " -F \'");
               strcat(OptionalParms, aFullName);
               strcat(OptionalParms, "\'");
            }

         }
      } /* if (rc == 0) */
      
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("f");										/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"f", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         aNameFrom = UnicodeToASCII(AttributeBuffer);
         if (aNameFrom[0] != '\0') {
            strcat(OptionalParms, " -f ");
            strcat(OptionalParms, aNameFrom);
         }
      } /* if (rc == 0) */
      
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("h");										/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"h", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         aHopCount = UnicodeToASCII(AttributeBuffer);
         if (aHopCount[0] != '\0') {
            strcat(OptionalParms, " -h ");
            strcat(OptionalParms, aHopCount);
         }
      } /* if (rc == 0) */
      
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      {																								/* 22/07/2001 */
         const CciChar *ucsAttr = CciString("X");										/* 22/07/2001 */
         rc = getAttribute(context,(CciChar *)ucsAttr, AttributeBuffer, sizeof(AttributeBuffer));/* 22/07/2001 */
         free((void *)ucsAttr) ;																/* 22/07/2001 */
      }																								/* 22/07/2001 */
#else																									/* 22/07/2001 */
      rc = getAttribute(context, L"X", AttributeBuffer, sizeof(AttributeBuffer));
#endif
		if (rc == 0)  {																			/* 22/07/2001 */
         aLogFileName = UnicodeToASCII(AttributeBuffer);
         if (aLogFileName[0] != '\0') {
            strcat(OptionalParms, " -X ");
            strcat(OptionalParms, aLogFileName);
         }
      } /* if (rc == 0) */

      p->SendmailOptionalParms = (char *)malloc(sizeof(char) * (strlen(OptionalParms) + 1 ));
      strcpy(p->SendmailOptionalParms, OptionalParms);

#ifdef _DEBUG
      fprintf(p->debugFile, "Using sendmail with optional sendmail parameters: %s\n", p->SendmailOptionalParms);
      fflush(p->debugFile);
      fprintf(p->debugFile, "-C = [%s]\n", aConfigFileName);
      fflush(p->debugFile);
      fprintf(p->debugFile, "-F = [%s]\n", aFullName);
      fflush(p->debugFile);
      fprintf(p->debugFile, "-f = [%s]\n", aNameFrom);
      fflush(p->debugFile);
      fprintf(p->debugFile, "-h = [%s]\n", aHopCount);
      fflush(p->debugFile);
      fprintf(p->debugFile, "-X = [%s]\n", aLogFileName);
      fflush(p->debugFile);
#endif
      free(aConfigFileName);
      free(aFullName);
      free(aNameFrom);
      free(aHopCount);
      free(aLogFileName);
   } /* if (p->SendmailOptionalParms == NULL) */

#endif /* #ifdef SENDMAIL */

   /* Get the root element */
   rootElement = cniRootElement(&rc, message);
   CCI_CHECK_RC();

   Element = cniFirstChild(&rc, rootElement);                                 /* 30/04/2001 */
   CCI_CHECK_RC();                                                            /* 30/04/2001 */

   while (Element) {                                                          /* 30/04/2001 */
      cniElementName(&rc, Element, (CciChar*)&elementName, sizeof(elementName));/* 30/04/2001 */
      CCI_CHECK_RC();                                                         /* 30/04/2001 */
#ifdef _DEBUG                                                                 /* 30/04/2001 */
      fprintf(p->debugFile, "Element Name (root tree) = %S\n", elementName);  /* 39/04/2001 */
      fflush(p->debugFile);                                                   /* 30/04/2001 */
#endif                                                                        /* 30/04/2001 */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if (CciCharCmp(elementName, (CciChar *)constMQMD)  == 0) {              /* 22/07/2001 */
#else                                                                         /* 30/04/2001 */
      if ( wcscmp(elementName, L"MQMD") == 0 ) {                              /* 30/04/2001 */
#endif                                                                        /* 30/04/2001 */
         mqmdFolder = Element;                                                /* 30/04/2001 */
         Element = 0;                                                         /* 30/04/2001 */
         MQMDvalid = TRUE;                                                    /* 15/05/2001 */
         continue;                                                            /* 30/04/2001 */
      }                                                                       /* 30/04/2001 */
      Element = cniNextSibling(&rc, Element);                                 /* 30/04/2001 */
      CCI_CHECK_RC();                                                         /* 30/04/2001 */
   } /* while (Element) */                                                    /* 30/04/2001 */


   if(MQMDvalid) {                                                            /* 15/05/2001 */
      Element = cniFirstChild(&rc, mqmdFolder);
      CCI_CHECK_RC();

      while (Element) {

         cniElementName(&rc, Element, (CciChar*)&elementName, sizeof(elementName));
         CCI_CHECK_RC();

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         if (CciCharCmp(elementName, (CciChar *)constPutDate) == 0) {			/* 22/07/2001 */
#else
         if ( wcscmp(elementName, L"PutDate") == 0 ) {
#endif
            valueDate = cniElementDateValue(&rc, Element);
#ifdef _DEBUG																						/* 21/04/2001 */
            fprintf(p->debugFile, "element PutDate found with rc = %d\n", rc);/* 21/04/2001 */
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif																								/* 21/04/2001 */
            /* CCI_CHECK_RC(); */                                             /* 15/05/2001 */
            if(rc != CCI_SUCCESS)                                             /* 15/05/2001 */
               MQMDvalid = FALSE;                                             /* 15/05/2001 */
         }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         if (CciCharCmp(elementName, (CciChar *)constPutTime) == 0) {			/* 22/07/2001 */
#else
         if ( wcscmp(elementName, L"PutTime") == 0 ) {
#endif
            valueTime = cniElementGmtTimeValue(&rc, Element);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "element PutTime found with rc = %d\n", rc);/* 21/04/2001 */
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            /* CCI_CHECK_RC(); */                                             /* 15/05/2001 */
            if(rc != CCI_SUCCESS)                                             /* 15/05/2001 */
               MQMDvalid = FALSE;                                             /* 15/05/2001 */
         }                                                                    /* 15/05/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
         if (CciCharCmp(elementName, (CciChar *)constMsgId) == 0) {				/* 22/07/2001 */
#else
         if ( wcscmp(elementName, L"MsgId") == 0 ) {
#endif
            //Create Blob Structure
		      Blob = (struct CciByteArray *)malloc(sizeof(struct CciByteArray));
            Blob->pointer = (char *)malloc(sizeof(char) * 1024);
            Blob->size    = 1024;
            cniElementByteArrayValue(&rc, Element, Blob);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "element MsgId   found with rc = %d\n", rc);/* 21/04/2001 */
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            /* CCI_CHECK_RC(); */                                             /* 15/05/2001 */
            if(rc != CCI_SUCCESS) {                                           /* 15/05/2001 */
               MQMDvalid = FALSE;                                             /* 15/05/2001 */
               MailMsgId[0] = '\0';                                           /* 15/05/2001 */
            }                                                                 /* 15/05/2001 */
            else {
               pTmp = Blob->pointer;
         
               MailMsgId[0] = '\0';

               Checksum = (unsigned char)0;                                   /* 19/10/2001 */

               for (i = 0 ; i < 24; i++) {
                  sprintf(pStr, "%.2X", (unsigned char)pTmp[i]);
                  Checksum = Checksum |(unsigned char)pTmp[i];                /* 19/10/2001 */
                  strcat(MailMsgId, pStr);
               }
#ifdef _DEBUG                                                                 /* 19/10/2001 */
               fprintf(p->debugFile, "element MsgId  Checksum = %d\n", Checksum);/* 19/10/2001 */
               fflush(p->debugFile);                                          /* 19/10/2001 */
#endif                                                                        /* 19/10/2001 */

               if (Checksum == (unsigned char)0)
                  MailMsgId[0] = '\0';
            }

            free(Blob->pointer); free(Blob);

         }

         Element = cniNextSibling(&rc, Element);
         CCI_CHECK_RC();

         } /* while (Element) */

   } /* if(MQMDvalid) */                                                      /* 15/05/2001 */

   if(MQMDvalid) {                                                            /* 15/05/2001 */
      sprintf(MailDate, "%02i %.3s %02i %02i:%02i:%02i GMT",                  /* 27/04/2001 */
              valueDate.day, MonthShortName(valueDate.month), valueDate.year, /* 27/04/2001 */
              valueTime.hour, valueTime.minute, (int)valueTime.second);       /* 27/04/2001 */
   }                                                                          /* 15/05/2001 */
   else {                                                                     /* 15/05/2001 */
      time(&ltime);                                                           /* 15/05/2001 */
      /* Obtain coordinated universal time: */                                /* 15/05/2001 */
      newtime = gmtime(&ltime);                                               /* 15/05/2001 */
      strftime(MailDate, sizeof(MailDate), "%d %b %Y %H:%M:%S GMT", newtime); /* 15/05/2001 */
   }                                                                          /* 15/05/2001 */
   
   /* Go to the last child of the root, the body <XML>*/
   Element = cniLastChild(&rc, rootElement);
   CCI_CHECK_RC();

   /* Go to the first child of the body <Message> */
   Element = cniFirstChild(&rc, Element);
   CCI_CHECK_RC();

   /* Go to the first child of the previous element, the level with data */
   Element = cniFirstChild(&rc, Element);
   CCI_CHECK_RC();

   valueChar = (CciChar *)malloc(sizeof(CciChar) * valueCharSize);            /* 21/04/2001 */

   /* Loop until all siblings are visited */
   while (Element) {

      elementNameSize = cniElementName(&rc, Element, (CciChar*)&elementName, sizeof(elementName));
      CCI_CHECK_RC();

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constFrom) == 0) || (CciCharCmp(elementName, (CciChar *)constMailFrom) == 0) ){/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"From") == 0 ) || (wcscmp(elementName, L"MailFrom") == 0) ){
#endif
         NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
         fprintf(p->debugFile, "element From    found with rc = %d\n", rc);   /* 21/04/2001 */
         fflush(p->debugFile);                                                /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
         if (rc == CCI_BUFFER_TOO_SMALL) {                                    /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            valueCharSize = NoOfBytes;                                        /* 21/04/2001 */
            valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
            NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
         }                                                                    /* 21/04/2001 */
         CCI_CHECK_RC();
         From = UnicodeToASCII(valueChar);
      }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constTo) == 0) || (CciCharCmp(elementName, (CciChar *)constMailTo) == 0) ){/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"To") == 0) || (wcscmp(elementName, L"MailTo") == 0) ){
#endif
         NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
         fprintf(p->debugFile, "element To      found with rc = %d\n", rc);   /* 21/04/2001 */
         fflush(p->debugFile);                                                /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
         if (rc == CCI_BUFFER_TOO_SMALL) {                                    /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            valueCharSize = NoOfBytes;                                        /* 21/04/2001 */
            valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
            NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
         }                                                                    /* 21/04/2001 */
         CCI_CHECK_RC();                                                      /* 21/04/2001 */
         aValueChar = UnicodeToASCII(valueChar);

         if (To == NULL) {
            To = NodeStringBufferAlloc();
            NodeStringBufferAdd(To, aValueChar);
         }
         else {
            NodeStringBufferAdd(To, "; ");
            NodeStringBufferAdd(To, aValueChar);
         }

         free(aValueChar);
      }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constCc) == 0) || (CciCharCmp(elementName, (CciChar *)constMailCc) == 0) ){/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"Cc") == 0) || (wcscmp(elementName, L"MailCc") == 0) ) {
#endif
         NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
         fprintf(p->debugFile, "element Cc      found with rc = %d\n", rc);   /* 21/04/2001 */
         fflush(p->debugFile);                                                /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
         if (rc == CCI_BUFFER_TOO_SMALL) {                                    /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            valueCharSize = NoOfBytes;                                        /* 21/04/2001 */
            valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
            NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
         }                                                                    /* 21/04/2001 */
         CCI_CHECK_RC();                                                      /* 21/04/2001 */
         aValueChar = UnicodeToASCII(valueChar);

         if (Cc == NULL) {
            Cc = NodeStringBufferAlloc();
            NodeStringBufferAdd(Cc, aValueChar);
         }
         else {
            NodeStringBufferAdd(Cc, "; ");
            NodeStringBufferAdd(Cc, aValueChar);
         }

         free(aValueChar);
      }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constBcc) == 0) || (CciCharCmp(elementName, (CciChar *)constMailBcc) == 0) ) {/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"Bcc") == 0) || (wcscmp(elementName, L"MailBcc") == 0) ) {
#endif
         NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
         fprintf(p->debugFile, "element Bcc     found with rc = %d\n", rc);   /* 21/04/2001 */
         fflush(p->debugFile);                                                /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
         if (rc == CCI_BUFFER_TOO_SMALL) {                                    /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            valueCharSize = NoOfBytes;                                        /* 21/04/2001 */
            valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
            NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
         }                                                                    /* 21/04/2001 */
         CCI_CHECK_RC();                                                      /* 21/04/2001 */
         aValueChar = UnicodeToASCII(valueChar);

         if (Bcc == NULL) {
            Bcc = NodeStringBufferAlloc();
            NodeStringBufferAdd(Bcc, aValueChar);
         }
         else {
            NodeStringBufferAdd(Bcc, "; ");
            NodeStringBufferAdd(Bcc, aValueChar);
         }

         free(aValueChar);
      }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constSubject) == 0) || (CciCharCmp(elementName, (CciChar *)constMailSubject) == 0) ) {/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"Subject") == 0) || (wcscmp(elementName, L"MailSubject") == 0) ) {
#endif
         NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
         fprintf(p->debugFile, "element subject found with rc = %d\n", rc);   /* 21/04/2001 */
         fflush(p->debugFile);                                                /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
         if (rc == CCI_BUFFER_TOO_SMALL) {                                    /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
            fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
            fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
            valueCharSize = NoOfBytes;                                        /* 21/04/2001 */
            valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
            NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
         }                                                                    /* 21/04/2001 */
         CCI_CHECK_RC();
         Subject = UnicodeToASCII(valueChar);
      }

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      if ( (CciCharCmp(elementName, (CciChar *)constBody) == 0) || (CciCharCmp(elementName, (CciChar *)constMailBody) == 0) ) {/* 22/07/2001 */
#else
      if ( (wcscmp(elementName, L"Body") == 0) || (wcscmp(elementName, L"MailBody") == 0) ) {
#endif
			MailBodyLine = cniFirstChild(&rc, Element);									/* 20/06/2001 */

			if(MailBodyLine) {																	/* 20/06/2001 */
				while(MailBodyLine) {															/* 20/06/2001 */
					NoOfBytes = cniElementCharacterValue(&rc, MailBodyLine, valueChar, valueCharSize);/* 20/06/2001 */
#ifdef _DEBUG                                                                 /* 20/06/2001 */
					fprintf(p->debugFile, "element Line    found with rc = %d\n", rc);/* 20/06/2001 */
					fflush(p->debugFile);                                          /* 20/06/2001 */
#endif																								/* 20/06/2001 */
					if (rc == CCI_BUFFER_TOO_SMALL) {                              /* 20/06/2001 */
#ifdef _DEBUG                                                                 /* 20/06/2001 */
						fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
						fflush(p->debugFile);                                       /* 20/06/2001 */
#endif                                                                        /* 20/06/2001 */
						valueCharSize = NoOfBytes;                                  /* 20/06/2001 */
						valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 20/06/2001 */
						NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 20/06/2001 */
					}                                                              /* 20/06/2001 */
					CCI_CHECK_RC();                                                /* 20/06/2001 */
					aValueChar = UnicodeToASCII(valueChar);								/* 20/06/2001 */

					if (Line == NULL) {															/* 20/06/2001 */
						Line = NodeStringBufferAlloc();										/* 20/06/2001 */
						NodeStringBufferAdd(Line, aValueChar);								/* 20/06/2001 */
					}																					/* 20/06/2001 */
					else {																			/* 20/06/2001 */
						NodeStringBufferAdd(Line, "\n");										/* 20/06/2001 */
						NodeStringBufferAdd(Line, aValueChar);								/* 20/06/2001 */
					}																					/* 20/06/2001 */

					free(aValueChar);																/* 20/06/2001 */

					MailBodyLine = cniNextSibling(&rc, MailBodyLine);					/* 20/06/2001 */
				}																						/* 20/06/2001 */
				Body = Line->buf;																	/* 20/06/2001 */
			}																							/* 20/06/2001 */
			else {																					/* 20/06/2001 */
				NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);
#ifdef _DEBUG                                                                 /* 21/04/2001 */
				fprintf(p->debugFile, "element Body    found with rc = %d\n", rc);/* 21/04/2001 */
				fflush(p->debugFile);                                             /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
				if (rc == CCI_BUFFER_TOO_SMALL) {                                 /* 21/04/2001 */
#ifdef _DEBUG                                                                 /* 21/04/2001 */
					fprintf(p->debugFile, "CCI_BUFFER_TOO_SMALL, NoOfBytes = %d, valueCharSize = %d\n", NoOfBytes, valueCharSize);
					fflush(p->debugFile);                                          /* 21/04/2001 */
#endif                                                                        /* 21/04/2001 */
					valueCharSize = NoOfBytes;                                     /* 21/04/2001 */
					valueChar = (CciChar *)realloc(valueChar, sizeof(CciChar) * valueCharSize);/* 21/04/2001 */
					NoOfBytes = cniElementCharacterValue(&rc, Element, valueChar, valueCharSize);/* 21/04/2001 */
				}                                                                  /* 21/04/2001 */
				CCI_CHECK_RC();
				Body = UnicodeToASCII(valueChar);
			}																							/* 20/06/2001 */
      }

      Element = cniNextSibling(&rc, Element);
   }

   free(valueChar);

   if (p->BasicBcc != NULL) {
      if (Bcc == NULL) {
         Bcc = NodeStringBufferAlloc();
         NodeStringBufferAdd(Bcc, p->BasicBcc);
      }
      else {
         NodeStringBufferAdd(Bcc, "; ");
         NodeStringBufferAdd(Bcc, p->BasicBcc);
      }
   } /* if (BasicMailBcc != NULL) */

#ifdef SMTP
   if(p->SMTPMailFrom != NULL) {
      if (From != NULL)
         free(From);
      From = p->SMTPMailFrom;
   } /* if(p->SMTPMailFrom != NULL) */
#endif

#ifdef SENDMAIL
   Parm0 = p->SendmailOptionalParms;
   Parm1 = (unsigned short)0;
#else
   Parm0 = p->SMTPHostName;
   Parm1 = p->SMTPHostPort == (unsigned short)0 ? (unsigned short)25 : p->SMTPHostPort;
#endif

#ifdef _DEBUG
   fprintf(p->debugFile, "From        %s\n", From);
   fprintf(p->debugFile, "To          %s\n", To  == NULL ? "" : To->buf);
   fprintf(p->debugFile, "Cc          %s\n", Cc  == NULL ? "" : Cc->buf);
   fprintf(p->debugFile, "Bcc         %s\n", Bcc == NULL ? "" : Bcc->buf);
   fprintf(p->debugFile, "Subject     %s\n", Subject);
   fprintf(p->debugFile, "Date        %s\n", MailDate);
   fprintf(p->debugFile, "MESSAGE-ID  %s\n", MailMsgId);
   fprintf(p->debugFile, "Body        %s\n", Body);
   fprintf(p->debugFile, "Priority    %d\n", p->BasicPriority);               /* 13/04/2001 */
   fprintf(p->debugFile, "Parm0 = [%s]\n", Parm0);
   fflush(p->debugFile);
#endif

   if ((To == NULL) && (Cc == NULL) && (Bcc == NULL)) {
      /* No recipients */
      rcTcp = -1;
      ErrorString = getErrorString(RC_NO_RECIPIENTS);
   }
   else {
      /* Try to send the e-mail message ..*/
      ErrorString = SendEmailMessage( Parm0,
                                      Parm1,                                  /* 28/04/2001 */
                                      From,
                                      To  == NULL ? "" : To->buf,
                                      Cc  == NULL ? "" : Cc->buf,
                                      Bcc == NULL ? "" : Bcc->buf,
                                      Subject,
                                      MailDate,
                                      MailMsgId,
                                      Body,
                                      p->BasicPriority,                       /* 13/04/2001 */
                                      &rcTcp );

#ifdef _DEBUG      
      fprintf (p->debugFile, "SendEmailMessage  returns rcTcp = %d\n", rcTcp);
      fprintf (p->debugFile, "ErrorString = %s\n", ErrorString);
      fflush(p->debugFile);
#endif

      if(To != NULL)
         NodeStringBufferFree(To);
      if(Cc != NULL)
         NodeStringBufferFree(Cc);
      if(Bcc != NULL)
         NodeStringBufferFree(Bcc);
   }

#ifdef SMTP
   if(p->SMTPMailFrom == NULL)
      if(From != NULL)
         free(From);
#else
   free(From);
#endif
   free(Subject);
	if(Line == NULL)																				/* 20/06/2001 */
		free(Body);
	else																								/* 20/06/2001 */
		NodeStringBufferFree(Line);															/* 20/06/2001 */


   /* propagate the message to the output terminal */
   if (rcTcp == 0) {
      /* get the handle to the output terminal */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constOut);/* 22/07/2001 */
#else																									/* 22/07/2001 */
      outputTerminalObject = getOutputTerminalHandle(context, L"out");
#endif																								/* 22/07/2001 */
   }
   else {
      /* Create a new message for output */
      outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));
      CCI_CHECK_RC();
      /* Get the root element of the output message */
      outRootElement = cniRootElement(&rc, outMsg);
      CCI_CHECK_RC();
      /* Copy the contents of the input message to the output message */
      cniCopyElementTree(&rc, rootElement, outRootElement);
      CCI_CHECK_RC();
      /* Get the last child of root (ie the body) and it's name */
      bodyChild = cniLastChild(&rc, outRootElement);
      CCI_CHECK_RC();
      /* Navigate to the root of the XML message */
      lastChild = cniLastChild(&rc, bodyChild);
      CCI_CHECK_RC();

      newElement = cniCreateElementAsLastChild(&rc, lastChild);
      CCI_CHECK_RC();
      cniSetElementType(&rc, newElement, CCI_ELEMENT_TYPE_NAME);
      CCI_CHECK_RC();
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		cniSetElementName(&rc, newElement, (CciChar *)constSendMailError);		/* 22/07/2001 */
#else																									/* 22/07/2001 */
      cniSetElementName(&rc, newElement, L"SendMailError");
#endif																								/* 22/07/2001 */
      CCI_CHECK_RC();

      newElement = cniCreateElementAsLastChild(&rc, newElement);
      CCI_CHECK_RC();
      cniSetElementType(&rc, newElement, CCI_ELEMENT_TYPE_VALUE);
      CCI_CHECK_RC();
      
      pErrorString = ASCIIToUnicode(ErrorString);
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		cniSetElementCharacterValue(&rc, newElement, (const CciChar *)pErrorString, CciCharLen((const CciChar *)pErrorString) );
#else																									/* 22/07/2001 */      
      cniSetElementCharacterValue(&rc, newElement, (const CciChar *)pErrorString, wcslen((const CciChar *)pErrorString) );
#endif
      CCI_CHECK_RC();
      
      free(pErrorString);

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constFailure);/* 22/07/2001 */
#else																									/* 22/07/2001 */
      outputTerminalObject = getOutputTerminalHandle(context, L"failure");
#endif																								/* 22/07/2001 */
      message = outMsg;
   }
   
   /* If the terminal exists and is attached, propagate to it */
   if (outputTerminalObject) {
      if (cniIsTerminalAttached(&rc, outputTerminalObject)) {
         if (rc == CCI_SUCCESS) {
            retvalue = cniPropagate(&rc, outputTerminalObject, destinationList, exceptionList, message);
            if (rc != CCI_SUCCESS) {
               if (rc == CCI_EXCEPTION) {
                  /* Get details of the exception */
                  memset(&exception_st, 0, sizeof(exception_st));
                  cciGetLastExceptionData(&rc, &exception_st);
                  /* Any local error handling may go here */
                  /* If the exception is "caught" it must be re-thrown */
                  cciRethrowLastException(0);
               } /* if (rc == CCI_EXCEPTION) */
               /* Some other error...log it */
               cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
                 (CciChar*)constMQSeriesIntegrator2, 0, "Propagate failed", 0);
#else
                 L"MQSIv201", 0, "Propagate failed", 0);
#endif
            } /* if (rc != CCI_SUCCESS) */
            else {
               /* Message was propagated successfully */
               cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
                 (CciChar*)constMQSeriesIntegrator2, 0, "Message propagated", 0);
#else 
                 L"MQSIv201", 0, "Message propagated", 0);
#endif 
            }
         } /* if (rc == CCI_SUCCESS) */
         else {
            if (rc == CCI_EXCEPTION) {
               cciRethrowLastException(0);
            }
            else {
               /* Some other error...log it */
               cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
                 (CciChar*)constMQSeriesIntegrator2, 0, "Error occurred", 0);
#else 
                 L"MQSIv201", 0, "Error occurred", 0);
#endif 
            }
         }
      }
   }
   else {
      /* Terminal did not exist...severe internal error */
      cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "", 
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
           (CciChar*)constMQSeriesIntegrator2, 0, "Terminal not connected", 0);
#else 
           L"MQSIv201", 0, "Terminal not connected", 0);
#endif 
   }

   if (rcTcp != 0) {
      /* Delete the message we created now we have finished with it */
      cniDeleteMessage(0, outMsg);
   }

#ifdef _DEBUG
   fprintf(p->debugFile, "<- %s\n", functionName);
   fflush(p->debugFile);
#ifdef WIN32
   DisplayMemoryStatus();
#endif
#endif

   return;
}

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initNodeConstants(void)
{
	constMQSeriesIntegrator2 = CciString(CONST_MQSERIES_INTEGRATOR_2);
	constPlugInNode          = CciString(CONST_PLUGIN_NODE);
	constPluginNodeFactory   = CciString(CONST_PLUGIN_NODE_FACTORY);

	constOut                 = CciString(CONST_OUT);
	constFailure             = CciString(CONST_FAILURE);

	constYes                 = CciString(CONST_YES);

	constMQMD                = CciString(CONST_MQMD);
	constPutDate             = CciString(CONST_PUTDATE);
	constPutTime             = CciString(CONST_PUTTIME);
	constMsgId	             = CciString(CONST_MSGID);

	constFrom	             = CciString(CONST_FROM);
	constMailFrom	          = CciString(CONST_MAILFROM);
	constTo						 = CciString(CONST_TO);
	constMailTo					 = CciString(CONST_MAILTO);
	constCc						 = CciString(CONST_CC);
	constMailCc					 = CciString(CONST_MAILCC);
	constBcc						 = CciString(CONST_BCC);
	constMailBcc				 = CciString(CONST_MAILBCC);
	constSubject				 = CciString(CONST_SUBJECT);
	constMailSubject			 = CciString(CONST_MAILSUBJECT);
	constBody					 = CciString(CONST_BODY);
	constMailBody				 = CciString(CONST_MAILBODY);
	constLine					 = CciString(CONST_LINE);
	constMailLine				 = CciString(CONST_MAILLINE);

	constHighest				 = CciString(CONST_HIGHEST);
	constHigh					 = CciString(CONST_HIGH);
	constNormal					 = CciString(CONST_NORMAL);
	constLow						 = CciString(CONST_LOW);
	constLowest					 = CciString(CONST_LOWEST);

	constSendMailError		 = CciString(CONST_SENDMAILERROR);
}
#endif

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                                       */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine.                                       */
/*                                                                            */
/* The responsibilities of the plug-in are to:                                */
/*  - create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/*  - define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/*  - define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/*  - define input and output terminals supported by each node.               */
/*  - return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
CciFactory LilFactoryExportPrefix *bipGetMessageflowNodeFactory()
{
   /*@unused@*/static char *functionName = (char *)"bipGetMessageflowNodeFactory()";
   CciFactory*       factoryObject;
   int               rc = 0;
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
#else
   CciChar           factoryName[] = L"SendMailPlugInNodeFactory";
#endif
   CCI_EXCEPTION_ST  exception_st;
   static CNI_VFT    vftable;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   /* Before we proceed we need to initialise all the static constants */
   /* that may be used by the plug-in.                                 */
   initNodeConstants();
#endif
   /* Create the Node Factory for this plug-in */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPluginNodeFactory);
#else
   factoryObject = cniCreateNodeFactory(0, factoryName);
#endif

   if (factoryObject == CCI_NULL_ADDR) {
      if (rc == CCI_EXCEPTION) {
         /* Get details of the exception */
         memset(&exception_st, 0, sizeof(exception_st));
         cciGetLastExceptionData(&rc, &exception_st);

         /* Any local error handling may go here */

         /* Rethrow the exception */
         cciRethrowLastException(&rc);
      }
      /* Any further local error handling can go here */
   }
   else {
      /* Setup function table with pointers to node implementation functions */
      memset(&vftable, 0, sizeof(vftable));
      vftable.iFpCreateNodeContext     = createNodeContext;
      vftable.iFpGetAttributeName      = getAttributeName;
      vftable.iFpSetAttribute          = setAttribute;
      vftable.iFpGetAttribute          = getAttribute;
      vftable.iFpEvaluate              = evaluate;

      /* Define a node type supported by our factory. If any errors/exceptions    */
      /* occur during the execution of this utility function, then as we have not */
      /* supplied the returnCode argument, the exception will bypass the plugin   */
      /* and be directly handled by the broker.                                   */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      cniDefineNodeClass(0, factoryObject, (CciChar*)constPlugInNode, &vftable);
#else                                  
      cniDefineNodeClass(&rc, factoryObject, L"SendMailPlugInNode", &vftable); 
#endif

      if (rc != CCI_SUCCESS) {
         /* Error Handling */
      }
   }

   /* Return address of this factory object to the broker */
   return(factoryObject);
}


static char *getErrorString(int rc)															/* 28/07/2001 */
{
   switch(rc) {
      case RC_NO_RECIPIENTS:
         return("No recipients found");
         break;
      default:
         return("Unknown error");
         break;
   }

	return("Unknown error");
}


static char GetPriorityNumber(CciChar *Priority)                              /* 13/04/2001 */
{
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	if (CciCharCmp((CciChar *)constHighest, Priority ) == 0)							/* 28/07/2001 */
		return (char)1;																			/* 28/07/2001 */
	if (CciCharCmp((CciChar *)constHigh, Priority ) == 0)								/* 28/07/2001 */
		return (char)2;																			/* 28/07/2001 */
	if (CciCharCmp((CciChar *)constNormal, Priority ) == 0)							/* 28/07/2001 */
		return (char)3;																			/* 28/07/2001 */
	if (CciCharCmp((CciChar *)constLow, Priority ) == 0)								/* 28/07/2001 */
		return (char)4;																			/* 28/07/2001 */
	if (CciCharCmp((CciChar *)constLowest, Priority ) == 0)							/* 28/07/2001 */
		return (char)5;																			/* 28/07/2001 */

	return (char)3;																				/* 22/07/2001 */
#else																									/* 22/07/2001 */
   static CciChar *PriorityArray[] = {
       L"Highest",
       L"High", 
       L"Normal", 
       L"Low", 
       L"Lowest",
       NULL,
    }; 

   int i = 3;

   for(i = 0; i < 5 &&  PriorityArray[i] != NULL; i++) {
      if(wcscmp(Priority, PriorityArray[i]) == 0)
         return(char)(i + 1);
   }

   return (char)3;
#endif																								/* 22/07/2001 */
}


