/******************************************************************************/
/*                                                                            */
/* Version 1.8                                                                */
/* Author: Arjan van Vught                 Date: September 2001               */
/*                                                                            */
/******************************************************************************/
/* Version 1.0   First release NT                                             */
/* Version 1.1   First release AIX                                            */
/* Version 1.2   Priority field added                                         */
/* Version 1.3   SMTP server port number added                                */
/* Version 1.4   Improved SMTP server error messages                          */
/* Version 1.5   Enhanced DEBUG info                                          */
/* Version 1.6   First release Sun Solaris                                    */
/* Version 1.7   First release HPUX                                           */
/* Version 1.7   First release Linux                                          */
/******************************************************************************/
/*                                                                            */
/* Note: This implementation is not necessarily portable                      */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/
/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "smtp.h"

/********************************************************************************************/
/* DEBUG */
#ifdef WIN32
#ifdef _DEBUG
   FILE *OST = NULL;

   #define DBGPRINTF(x)  {\
   OST = fopen("\\DebugSMTPClient.txt","a+"); \
      if(OST) fprintf x; \
         fclose(OST); \
   }
#endif
#endif

#if defined (AIX) || defined(SUN) || defined(HPUX)	|| defined (LINUX)			/* 16/09/2001 */
#ifdef _DEBUG
   FILE *OST = NULL;

   #define DBGPRINTF(x)  {\
   OST = fopen("/tmp/DebugSMTPClient.txt","a+"); \
      if(OST) fprintf x; \
         fclose(OST); \
   }
#endif
#endif

#ifdef _DEBUG                                                                 /* 27/04/2001 */
void DebugTime(void)                                                          /* 27/04/2001 */
{                                                                             /* 27/04/2001 */
   struct tm *newtime = NULL;                                                 /* 27/04/2001 */
   time_t aclock;                                                             /* 27/04/2001 */
   time(&aclock);                                                             /* 27/04/2001 */
   newtime = localtime(&aclock);                                              /* 27/04/2001 */
   DBGPRINTF((OST, "===============================\n"));                     /* 27/04/2001 */
   DBGPRINTF((OST, "= The current date and time are: %s", asctime(newtime))); /* 27/04/2001 */
   DBGPRINTF((OST, "===============================\n"));                     /* 27/04/2001 */
}                                                                             /* 27/04/2001 */
#endif                                                                        /* 27/04/2001 */
/* DEBUG */
/********************************************************************************************/


struct S_SmtpProto SmtpProto[] = 
{
  { NULL,  /*  CONNECTION  */
     {  { "220", SMTP_SUCCESS,           },
        { "421", SMTP_SERVICECLOSED,     },
        {  NULL,  0                        } }  },
  { "HELO %s",
     {  { "250", SMTP_SUCCESS,           },
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "500", SMTP_SYNTAXERROR,       },
        { "501", SMTP_SYNTAXERROR,       },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "550", SMTP_SERVERCANTEXECUTE, },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "MAIL FROM:<%s>",
     {  { "250", SMTP_SUCCESS,           },
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "451", SMTP_SERVERCANTEXECUTE, },
        { "452", SMTP_STORAGEEXCEDED,    },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { "501", SMTP_SERVERCANTEXECUTE, },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "552", SMTP_STORAGEEXCEDED,    },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "MAIL FROM:<%s> SIZE=%ld",
     {  { "250", SMTP_SUCCESS,           },
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "451", SMTP_SERVERCANTEXECUTE, },
        { "452", SMTP_STORAGEEXCEDED,    },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { "501", SMTP_SERVERCANTEXECUTE, },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "552", SMTP_STORAGEEXCEDED,    },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "RCPT TO: ",
     {  { "250", SMTP_SUCCESS,           },
        { "251", SMTP_SUCCESS,           }, /* forwarded */
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "451", SMTP_SERVERCANTEXECUTE, },
        { "452", SMTP_STORAGEEXCEDED,    },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { "501", SMTP_SERVERCANTEXECUTE, },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "551", SMTP_UNKNOWNUSER,       },
        { "552", SMTP_STORAGEEXCEDED,    },
        { "553", SMTP_SERVERCANTEXECUTE, },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "DATA",
     {  { "354", SMTP_SUCCESS,           },
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "451", SMTP_SERVERCANTEXECUTE, },
        { "452", SMTP_STORAGEEXCEDED,    },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "554", SMTP_SERVERCANTEXECUTE, },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "\r\n.",  /*  END DATA  */
     {  { "250", SMTP_SUCCESS,           },
        { "451", SMTP_SERVERCANTEXECUTE, },
        { "452", SMTP_STORAGEEXCEDED,    },
        { "552", SMTP_STORAGEEXCEDED,    },
        { "554", SMTP_SERVERCANTEXECUTE, },
        { NULL,  0                         } }   },
   { "VRFY %s",
     {  { "250", SMTP_SUCCESS,           },
        { "251", SMTP_FORWARDED          },
        { "421", SMTP_SERVERCANTEXECUTE, },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { "501", SMTP_SERVERCANTEXECUTE, },
        { "504", SMTP_NOTIMPLEMENTED,    },
        { "550", SMTP_UNKNOWNUSER,       },
        { "551", SMTP_UNKNOWNUSER,       },
        { "553", SMTP_SERVERCANTEXECUTE, },
        { "555", SMTP_SYNTAXERROR,       },
        { NULL,  0                         } }   },
   { "QUIT",
     {  { "221", SMTP_SUCCESS,           },
        { "250", SMTP_SUCCESS,           },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { NULL,  0                         } }   },
   { "REST",
     {  { "250", SMTP_SUCCESS,           },
        { "500", SMTP_SERVERCANTEXECUTE, },
        { "501", SMTP_SERVERCANTEXECUTE, },
        { "503", SMTP_SERVERCANTEXECUTE, },
        { "504", SMTP_SERVERCANTEXECUTE, },
        { NULL,  0                         } }   },
}; /* SMTP protocol */



int SmtpSendMessage(SMTPINFO	*SMTPinfo,
                    char		*MailFrom, 
                    char		*MailTo, 
                    char		*MailCc, 
                    char		*MailBcc, 
                    char		*MailSubject,
                    char		*MailDate,
                    char		*MailMsgId,
                    char		*MailBody,
                    char		 MailPriority,
						  char	  **Rbuffer)
{
   SOCKET S;

   int   rc    =  0;
   int   i     =  0;

   static char R[10000];
   char Buffer[2048];

   char *MailToRcpt  = NULL;
   char *MailFromCmd = NULL;                                                  /* 03/03/2001 */

#ifdef _DEBUG                                                                 /* 27/04/2001 */
   DebugTime();                                                               /* 27/04/2001 */
   DBGPRINTF((OST, "--> SmtpSendMessage\n"));                                 /* 27/04/2001 */
	DBGPRINTF((OST, "--- MailFrom   <%s>\n", MailFrom));                       /* 07/07/2001 */
   DBGPRINTF((OST, "--- MailTo     <%s>\n", MailTo));                         /* 07/07/2001 */
	DBGPRINTF((OST, "--- MailCc     <%s>\n", MailCc));                         /* 07/07/2001 */
	DBGPRINTF((OST, "--- MailBcc    <%s>\n", MailBcc));                        /* 07/07/2001 */
#endif                                                                        /* 27/04/2001 */

   if ( ( MailToRcpt = (char *)malloc(strlen(MailTo) + strlen(MailCc) + strlen(MailBcc) + 2) ) == NULL ) {
#ifdef _DEBUG                                                                 /* 27/04/2001 */
      DBGPRINTF((OST, "<-- SmtpSendMessage - TCP_INSMEMORY\n"));              /* 27/04/2001 */
#endif                                                                        /* 27/04/2001 */
		*Rbuffer = TCPIPErrorString(TCP_INSMEMORY);										/* 25/06/2001 */
      return TCP_INSMEMORY;
   }

   /* 03/03/2001 */
   if ( ( MailFromCmd = (char *)malloc(strlen(MailFrom) + 1) ) == NULL ) {
#ifdef _DEBUG                                                                 /* 27/04/2001 */
      DBGPRINTF((OST, "<-- SmtpSendMessage - TCP_INSMEMORY\n"));              /* 27/04/2001 */
#endif                                                                        /* 27/04/2001 */
		*Rbuffer = TCPIPErrorString(TCP_INSMEMORY);										/* 25/06/2001 */
      return TCP_INSMEMORY;
   }

   strcpy(MailToRcpt, MailTo);

	if (MailCc[0] != '\0') {																	/* 07/07/2001 */
	   strcat(MailToRcpt, ";");
		strcat(MailToRcpt, MailCc);
	}																									/* 07/07/2001 */

	if (MailBcc[0] != '\0') {																	/* 07/07/2001 */
		strcat(MailToRcpt, ";");
		strcat(MailToRcpt, MailBcc);
	}																									/* 07/07/2001 */

   memset(R, 0, sizeof(R));																	/* 07/07/2001 */
	*Rbuffer = R;

   if ( ( rc = TcpConnect(&S, SMTPinfo) ) != TCP_OK ){
#ifdef _DEBUG
      DBGPRINTF((OST, "<-- SmtpSendMessage - TcpConnect failed\n"));          /* 27/04/2001 */
#endif
		*Rbuffer = TCPIPErrorString(rc);														/* 25/06/2001 */
      return rc;
   }

   rc = TcpRecv(S,R,10000);
   if ( ( rc = SmtpGetRc(SmtpProto[_SMTP_CONNECT], R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "<-- SmtpSendMessage - _SMTP_CONNECT failed\n"));       /* 27/04/2001 */
#endif
		*Rbuffer = TCPIPErrorString(rc);														/* 25/06/2001 */
      return rc;
   }

   /* sends the HELO command */
   sprintf(Buffer, SmtpProto[_SMTP_HELO].Cmd, SMTPinfo->HostName);

   rc = SmtpTcpSend(S, Buffer, strlen(Buffer));
   rc = TcpRecv(S,R,10000);

   if ( ( rc = SmtpGetRc(SmtpProto[_SMTP_HELO], R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- _SMTP_HELO failed\n"));
#endif
		return rc;
   }

   /* sends the MAILFROM command */
   if (MailFrom[strlen(MailFrom) - 1] == '>') {                               /* 03/03/2001 */
      strcpy(MailFromCmd, MailFrom);                                          /* 03/03/2001 */
      MailFromCmd[strlen(MailFromCmd) - 1] = '\0';                            /* 03/03/2001 */
      i = 0;                                                                  /* 03/03/2001 */
      while( (MailFromCmd[i] != '\0') && (MailFromCmd[i] != '<') )            /* 03/03/2001 */
         i++;                                                                 /* 03/03/2001 */
      sprintf(Buffer, SmtpProto[_SMTP_MAILFROM].Cmd, MailFromCmd + i + 1);    /* 03/03/2001 */
      free(MailFromCmd);                                                      /* 03/03/2001 */
   }                                                                          /* 03/03/2001 */
   else                                                                       /* 03/03/2001 */
      sprintf(Buffer, SmtpProto[_SMTP_MAILFROM].Cmd, MailFrom);
   
   rc = SmtpTcpSend(S, Buffer, strlen(Buffer));
   rc = TcpRecv(S,R,10000);

   if ( ( rc = SmtpGetRc(SmtpProto[_SMTP_MAILFROM], R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- _SMTP_MAILFROM failed\n"));
#endif
      return rc;
   }

   /* sends the RCPT TO command */
   if ( ( rc = SmtpSendRcpt(S, MailToRcpt, R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- Send RCPT TO failed\n"));
#endif
      return rc;
   }

#ifndef _DEBUG
   free(MailToRcpt);
#endif
   
   /* sends the DATA command */
   rc = SmtpTcpSend(S, SmtpProto[_SMTP_DATA].Cmd, strlen(SmtpProto[_SMTP_DATA].Cmd));
   rc = TcpRecv(S,R,10000);

   if ( ( rc = SmtpGetRc(SmtpProto[_SMTP_DATA], R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- _SMTP_DATA failed, rc = %d\n", rc));
#endif
      return rc;
   }

#ifdef _DEBUG
	DBGPRINTF((OST, "-- _SMTP_DATA ok\n"));
#endif

   /* sends the TO field */
   if (MailTo[0] != '\0')
      rc = SmtpSendTo(S, MailTo);

#ifdef _DEBUG
   DBGPRINTF((OST, "-- TO field, rc = %d\n", rc));
#endif

   /* sends the CC field */
   if (MailCc[0] != '\0')
      rc = SmtpSendCc(S, MailCc);

#ifdef _DEBUG
   DBGPRINTF((OST, "-- CC field, rc = %d\n", rc));
#endif

   /* sends the FROM field */
   sprintf(Buffer,"From: %s", MailFrom);
   rc = SmtpTcpSend(S, Buffer, strlen(Buffer));

#ifdef _DEBUG
   DBGPRINTF((OST, "-- FROM field, rc = %d\n", rc));
#endif
       
   /*  sends the DATE field */
   sprintf(R, "Date: %s\r\n", MailDate);
   rc = TcpSend(S, R,strlen(R));
       
   /*  sends the MESSAGE_ID field */
	if (MailMsgId[0] != '\0') {																/* 08/07/2001 */
		sprintf(R, "Message-ID: %s\r\n", MailMsgId);
		rc = TcpSend(S, R,strlen(R));
	}																									/* 08/07/2001 */
   
   /* sends the SUBJECT field */
   sprintf(R,"Subject: %s\r\n", MailSubject);
   rc = TcpSend(S, R,strlen(R));

   /* sends the PRIORITY field */
   rc = SmtpSendPriority(S, MailPriority);                                    /* 09/04/2001 */

   /* sends the IMPORTANCE field */
   strcpy(R,"Importance: Normal\r\n");
   rc = TcpSend(S, R,strlen(R));

   /* sends the REPLY-TO field */
   sprintf(Buffer,"Reply-to: %s", MailFrom);
   rc = SmtpTcpSend(S, Buffer, strlen(Buffer));

   /* sends the X-Priority field                                                 09/04/2001 */
   rc = SmtpSendXPriority(S, MailPriority);                                   /* 09/04/2001 */

   /* sends the X-MAILER field */
   sprintf(Buffer,"X-mailer: %s", XMAILER);
   rc = SmtpTcpSend(S, Buffer, strlen(Buffer));

   /* sends the blank line                                                       26/04/2001 */
   strcpy(R,"\r\n");                                                          /* 26/04/2001 */
   rc = TcpSend(S, R,strlen(R));                                              /* 26/04/2001 */

   /* sends the BODY field */
   rc = SmtpTcpSend(S, MailBody, strlen(MailBody));
   
   /* sends the ENDOFDATA command */
   rc = SmtpTcpSend(S, SmtpProto[_SMTP_ENDOFDATA].Cmd, strlen(SmtpProto[_SMTP_ENDOFDATA].Cmd));
   rc = TcpRecv(S,R,10000);

   if ( ( rc = SmtpGetRc(SmtpProto[_SMTP_ENDOFDATA], R) ) != SMTP_SUCCESS ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- _SMTP_ENDOFDATA failed\n"));
#endif
      return rc;
   }

   /* sends the QUIT command */
   rc = SmtpTcpSend(S, SmtpProto[_SMTP_QUIT].Cmd, strlen(SmtpProto[_SMTP_QUIT].Cmd));
   rc = TcpRecv(S,R,10000);
   rc = SmtpGetRc(SmtpProto[_SMTP_QUIT], R);

   if ( (rc = TcpClose(S)) != TCP_OK )
      return rc;

   return rc;
}


int SmtpGetRc(struct S_SmtpProto SmtpProto, char *Buffer)
{
   int Index = 0;
#ifdef _DEBUG
   DBGPRINTF((OST, "-> SmtpGetRc, buffer = %s\n", Buffer));
#endif

	/* translate the returned string */
   for(Index = 0; Index < SizeOfTab(SmtpProto.Answer) &&  SmtpProto.Answer[Index].String != NULL; Index++) {
      if(memcmp(Buffer, SmtpProto.Answer[Index].String, strlen(SmtpProto.Answer[Index].String)) == 0) {
#ifdef _DEBUG
         DBGPRINTF((OST, "- String %s Code %d\n", SmtpProto.Answer[Index].String, SmtpProto.Answer[Index].Code));
         DBGPRINTF((OST, "<- SmtpGetRc, rc = %d\n", SmtpProto.Answer[Index].Code));
#endif
         return(SmtpProto.Answer[Index].Code);
      }
   }

#ifdef _DEBUG
   DBGPRINTF((OST, "<- SmtpGetRc, rc = -1\n"));
#endif

   return -1;
}


int SmtpTcpSend(SOCKET S, char *Buffer, int BufferSize)
{
   char *p;
   int rc = 0;

#ifdef _DEBUG
   DBGPRINTF((OST, "-> SmtpTcpSend, Buffer = %.80s, BufferSize = %d\n", Buffer, BufferSize));
#endif

   p = (char *)malloc(sizeof(char) * (1 + BufferSize + strlen("\r\n") ) );

   if(p == NULL)
      return(TCP_INSMEMORY);

   /* sprintf(p, "%s\r\n", Buffer); */
   sprintf(p, "%.*s\r\n", BufferSize, Buffer);											/* 07/07/2001 */

   rc = TcpSend(S, p, BufferSize + strlen("\r\n"));

   free(p);

#ifdef _DEBUG
   DBGPRINTF((OST, "<- SmtpTcpSend, rc = %d\n", rc));
#endif

   return(rc);
}


int SmtpSendRcpt (SOCKET S, char *pTo, char *R)
{
   int   rc  = SMTP_SUCCESS;
   int   Index;

   char *p = NULL;
   char *Cmd;
   char *To = pTo;

#ifdef _DEBUG
   DBGPRINTF((OST, "-> SmtpSendRcpt, pTo = <%s>\n", pTo));
#endif

   p = (char *)malloc(sizeof(char) * ( 1 + strlen(To) + strlen(SmtpProto[_SMTP_RCPTTO].Cmd) ) );

   if (p == NULL)
      return(TCP_INSMEMORY);

#ifdef _DEBUG
		DBGPRINTF((OST, "-- Memory allocated\n"));
#endif

   while (*To != 0  &&  rc == SMTP_SUCCESS) {

#ifdef _DEBUG
		DBGPRINTF((OST, "-- Constructs the beginning of the command\n"));
#endif
      strcpy (p, SmtpProto[_SMTP_RCPTTO].Cmd);
      Cmd = p + strlen(SmtpProto[_SMTP_RCPTTO].Cmd);
    
#ifdef _DEBUG
		DBGPRINTF((OST, "-- Get the first recipient from the To line\n"));
#endif
      for (Index = 0; To[Index] != 0 && To[Index] != SMTP_SEPARATOR; Index++)
         Cmd[Index] = To[Index];

      Cmd[Index] = 0;

#ifdef _DEBUG
		DBGPRINTF((OST, "-- Skip spaces and affect To the next recipient\n"));
#endif
      while (To[Index] == SMTP_SEPARATOR || To[Index] == ' ' || To[Index]=='\t')
         Index++;

      To += Index;

      rc = SmtpTcpSend(S, p, strlen(p));
#ifdef _DEBUG
      DBGPRINTF((OST, "- RCPT TO Cmd: %s\n", p));
#endif
      TcpRecv(S,R,10000);
      rc = SmtpGetRc(SmtpProto[_SMTP_RCPTTO], R);


   } /* until either error or no more recipient */

   free(p);

#ifdef _DEBUG
   DBGPRINTF((OST, "<- SmtpSendRcpt, rc = %d\n", rc));
#endif
   
   return(rc);
}


int SmtpSendTo(SOCKET S, char *pTo)
{
   int   rc  = SMTP_SUCCESS;
   int   Index;

   char *p = NULL;
   char *To = pTo;
   char *Cmd;

#ifdef _DEBUG
   DBGPRINTF((OST, "-> SmtpSendTo\n"));
#endif

   p = (char *)malloc(sizeof(char) * (1 + strlen(To) + strlen("To:") ) );

   if (p == NULL)
      return(TCP_INSMEMORY);

   while (*To != 0  &&  rc == SMTP_SUCCESS) {

      /* constructs the beginning of the command */
      strcpy (p, "To:");
      Cmd = p + strlen("To:");


      /* get the first recipient from the To line */
      for (Index = 0; To[Index] != 0 && To[Index] != SMTP_SEPARATOR; Index++)
         Cmd[Index] = To[Index];

      Cmd[Index] = 0;

      /* skip spaces and affect To the next recipient */
      while (To[Index] == SMTP_SEPARATOR || To[Index] == ' ' || To[Index]=='\t')
         Index++;

      To += Index;

      rc = SmtpTcpSend(S, p, strlen(p));
#ifdef _DEBUG
      DBGPRINTF((OST, "- DATA To Cmd: %s\n", p));
#endif
      
   }

   free(p);

#ifdef _DEBUG
   DBGPRINTF((OST, "<- SmtpSendTo, rc = %d\n", rc));
#endif
   
   return(rc);
}


int SmtpSendCc(SOCKET S, char *pCc)
{
   int   rc  = SMTP_SUCCESS;
   int   Index;

   char *p = NULL;
   char *Cc = pCc;
   char *Cmd;

#ifdef _DEBUG
   DBGPRINTF((OST, "-> SmtpSendCc\n"));
#endif

   p = (char *)malloc(sizeof(char) * (1 + strlen(Cc) + strlen("Cc:") ) );

   if (p == NULL)
      return(TCP_INSMEMORY);

   while (*Cc != 0  &&  rc == SMTP_SUCCESS) {

      /* constructs the beginning of the command */
      strcpy (p, "Cc:");
      Cmd = p + strlen("Cc:");


      /* get the first recipient from the Cc line */
      for (Index = 0; Cc[Index] != 0 && Cc[Index] != SMTP_SEPARATOR; Index++)
         Cmd[Index] = Cc[Index];

      Cmd[Index] = 0;

      /* skip spaces and affect Cc the next recipient */
      while (Cc[Index] == SMTP_SEPARATOR || Cc[Index] == ' ' || Cc[Index]=='\t')
         Index++;

      Cc += Index;

      rc = SmtpTcpSend(S, p, strlen(p));
#ifdef _DEBUG
      DBGPRINTF((OST, "- DATA Cc Cmd: %s\n", p));
#endif
      
   }

   free(p);

#ifdef _DEBUG
   DBGPRINTF((OST, "<- SmtpSendCc, rc = %d\n", rc));
#endif
   
   return(rc);
}


char *SMTPErrorString(int rc)
{
   switch(rc) {
      case TCP_ERROR:
         return("TCP/IP error");
         break;
      case TCP_BADPORT:
         return("TCP/IP bad port");
         break;
      case TCP_HOSTUNKNOWN:
         return("TCP/IP host unknown");
         break;
      case TCP_NOMORESOCKET:
         return("TCP/IP no more sockets");
         break;
      case TCP_CONNECTFAILED:
         return("TCP/IP connection failed");
         break;
      case TCP_INSMEMORY:
         return("TCP/IP memory allocation error");
         break;
      case SMTP_UNEXPECTEDANSWER:
         return("Answer was not expected");
         break;
      case SMTP_SERVICECLOSED:
         return("Service unavailable");
         break;
      case SMTP_NOTIMPLEMENTED:
         return("Host recognize but can't exec cmd");
         break;
      case SMTP_MIMENOTSUPPORTED:
         return("Server doesn't support MIME ext");
         break;
      case SMTP_SERVERCANTEXECUTE:
         return("Refused by server");
         break;
      case SMTP_CANTCONNECT:
         return("Can not connect to the server");
         break;
      case SMTP_DATAERROR:
         return("Error during communication");
         break;
      case SMTP_SYNTAXERROR:
         return("Bad parameters");
         break;
      case SMTP_STORAGEEXCEDED:
         return("Server limits exceeded");
         break;
      case SMTP_UNKNOWNUSER:
         return("Unknown destinee");
         break;
      case SMTP_FORWARDED:
         return("Address OK,unknown on this server");
         break;
      default:
        return("Unknown SMTP error");
        break;
   }
}

/* TCP related functions */


char *TCPIPErrorString(int rc)																/* 25/06/2001 */
{
   switch(rc) {
      case TCP_ERROR:
         return("TCP/IP error");
         break;
      case TCP_BADPORT:
         return("TCP/IP bad port");
         break;
      case TCP_HOSTUNKNOWN:
         return("TCP/IP host unknown");
         break;
      case TCP_NOMORESOCKET:
         return("TCP/IP no more sockets");
         break;
      case TCP_CONNECTFAILED:
         return("TCP/IP connection failed");
         break;
      case TCP_INSMEMORY:
         return("TCP/IP memory allocation error");
         break;
      default:
        return("Unknown TCP/IP error");
        break;
   }
}

/* AIX SUN HPUX LINUX */
#if defined (AIX) || defined(SUN) || defined(HPUX)	|| defined (LINUX)			/* 16/09/2001 */
int closesocket(SOCKET S)
{
   return close(S);																				/* 19/07/2001 */
}
#endif


/*  Windows NT/2000 */
#ifdef WIN32
int TcpInit(void)
{
   int rc = -1;

   WSADATA W;

#ifdef _DEBUG
   DBGPRINTF((OST, "--> TcpInit\n"));
#endif

   rc = WSAStartup (0x101, &W);

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- TcpInit\n"));
#endif

	return(rc == 0 ? TCP_OK : TCP_ERROR);
}


int TcpCleanUp(void)
{
   int rc = TCP_ERROR;

#ifdef _DEBUG
   DBGPRINTF((OST, "--> TcpCleanUp\n"));
#endif

   rc = WSACleanup();

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- TcpCleanUp\n"));
#endif

   return(rc == 0 ? TCP_OK : TCP_ERROR);
}
#endif


/* AIX SUN */
#if defined (AIX) || defined(SUN) || defined(HPUX)	|| defined (LINUX)			/* 16/09/2001 */
int TcpInit(void)
{
#ifdef _DEBUG
   DBGPRINTF((OST, "<--> TcpInit\n"));
#endif
   return(TCP_OK);
}


int TcpCleanUp(void)
{ 
#ifdef _DEBUG
   DBGPRINTF((OST, "<--> TcpCleanUp\n"));
#endif
   return(TCP_OK);
}
#endif


int TcpConnect(SOCKET *S, SMTPINFO *SMTPinfo)
{
   
   int i = 0;
#ifdef _DEBUG
   DBGPRINTF((OST, "-> TcpConnect\n"));
#endif

   *S = INVALID_SOCKET;

   if ( (*S = socket(AF_INET, SOCK_STREAM,0) ) == SOCKET_ERROR ) {
#ifdef _DEBUG
      DBGPRINTF((OST, "<- TcpConnect : TCP_NOMORESOCKET\n"));
#endif
      return TCP_NOMORESOCKET;
   }

#ifdef WIN32  
   if ( (i = connect(*S, (const struct sockaddr *)&SMTPinfo->A, sizeof(SMTPinfo->A)) == SOCKET_ERROR)) {
#endif
#if defined (AIX) || defined(SUN) || defined(HPUX)	|| defined (LINUX)			/* 16/09/2001 */
   if ( (i = connect(*S, (const struct sockaddr *)&SMTPinfo->A, sizeof(SMTPinfo->A)) != 0)) {
#endif
      closesocket(*S);
#ifdef _DEBUG
      DBGPRINTF((OST, "<- TcpConnect : TCP_CONNECTFAILED\n"));
#endif
      return TCP_CONNECTFAILED;
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<- TcpConnect - connected to host %s on socket %d\n", inet_ntoa(SMTPinfo->A.sin_addr), *S));
#endif

   return TCP_OK;
}


int TcpClose(SOCKET S)
{
   int rc = TCP_ERROR;
#ifdef _DEBUG
   DBGPRINTF((OST, "--> TcpClose\n"));
#endif

   rc = closesocket(S);

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- TcpClose, rc = %d\n", rc == 0 ? TCP_OK : TCP_ERROR));
#endif

   return(rc == 0 ? TCP_OK : TCP_ERROR);
}


int TcpSend(SOCKET S, char *Buffer, int BufferSize)
{
   int      BytesSend      = 0;
   int      TotalBytesSend = 0;

#ifdef _DEBUG
   DBGPRINTF((OST, "--> TcpSend, BufferSize = %d\n", BufferSize));
#endif

   if (S == INVALID_SOCKET)
     return TCP_ERROR;
  
   for ( TotalBytesSend = 0, BytesSend = 1 ; TotalBytesSend < BufferSize  &&  BytesSend > 0 ;  TotalBytesSend += BytesSend) {
#ifdef _DEBUG
      DBGPRINTF((OST, "-- send on socket %d. %d bytes to be sent\n", S, BufferSize - TotalBytesSend));
#endif
      BytesSend = send (S, &Buffer[TotalBytesSend], BufferSize - TotalBytesSend, 0);

      if (BytesSend <= 0)
         printf("error send on socket %d\n", S);
   }

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- TcpSend, rc = %d\n", TotalBytesSend >= BufferSize ? TCP_OK : TCP_ERROR));
#endif

   return(TotalBytesSend >= BufferSize ? TCP_OK : TCP_ERROR);
}


int TcpRecv(SOCKET S, char *Buffer, int BufferSize)
{
   int   rc = 0;
#ifdef _DEBUG
   DBGPRINTF((OST, "--> TcpRecv, BufferSize = %d\n", BufferSize));
#endif

   memset(Buffer, 0, BufferSize);
   
   rc = recv(S, Buffer, BufferSize, 0);

#ifdef _DEBUG
   if (rc > 0) {
      DBGPRINTF((OST, "-- recv %d : %s\n", rc, Buffer));
   }
   else {
      DBGPRINTF((OST, "-- error receive on socket\n"));
   }

   DBGPRINTF((OST, "<-- TcpRecv, rc = %d\n", rc <= 0 ? TCP_ERROR : TCP_OK));
#endif

   return(rc <= 0 ? TCP_ERROR : TCP_OK );
}


int SmtpSendPriority(SOCKET S, char Priority)                                /* 09/04/2001 */
{

   static char R[64];
#ifdef _DEBUG
   DBGPRINTF((OST, "--> SmtpSendPriority, Priority = %d\n", Priority));
#endif

   memset(R, 0, sizeof(R));

   if (Priority < 1 || Priority > 5 || Priority == 3) {
      strncpy(R, "Priority: Normal\r\n", sizeof(R));
   }
   else if (Priority == 1 || Priority == 2) {
      strncpy(R, "Priority: High\r\n", sizeof(R));
   }
   else if (Priority == 4 || Priority == 5) {
      strncpy(R, "Priority: Low\r\n", sizeof(R));
   }

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- SmtpSendPriority\n"));
#endif

   return(TcpSend(S, R,strlen(R)));
}


int SmtpSendXPriority(SOCKET S, char Priority)                                /* 09/04/2001 */
{

   static char R[64];
#ifdef _DEBUG
   DBGPRINTF((OST, "--> SmtpSendXPriority, Priority = %d\n", Priority));
#endif

   memset(R, 0, sizeof(R));

   if (Priority < 1 || Priority > 5) {
      strncpy(R, "X-Priority: 3 (Normal)\r\n", sizeof(R));
   }
   else {
      sprintf(R,"X-Priority: %d (%s)\r\n", Priority, GetPriorityName(Priority));
   }

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- SmtpSendXPriority\n"));
#endif

   return(TcpSend(S, R,strlen(R)));
}


char *GetPriorityName(char Priority)                                          /* 09/04/2001 */
{
   static char *PriorityArray[] = {
       "      ",
       "Higest",
       "High  ", 
       "Normal", 
       "Low   ", 
       "Lowest",

    }; 

    return(Priority < 1 || Priority > 5) ? PriorityArray[3] : PriorityArray[Priority];
}


SMTPINFO *SmtpInit(char *SMTPhost, unsigned short SMTPhostPort, int *Reason)
{
   SMTPINFO *SMTPinit = NULL;

#ifdef _DEBUG
	DebugTime();               
   DBGPRINTF((OST, "-> SmtpInit\n"));
#endif

	/* The Windows TcpInit() function initiates use of WS2_32.DLL by a process */
	/* The AIX TcpInit() function is empty                                     */
   if ( TcpInit() == TCP_ERROR) {
#ifdef _DEBUG
	   DBGPRINTF((OST, "<-- SmtpInit - TcpInit failed\n"));    
#endif
		*Reason = TCP_ERROR;
      return NULL;
   }

	if ( (SMTPinit = (SMTPINFO *)malloc(sizeof(SMTPINFO))) == NULL) {
		*Reason = TCP_INSMEMORY;
		return NULL;
	}

	/* gethostname function returns the standard host name for the local machine */
   gethostname(SMTPinit->HostName, sizeof(SMTPinit->HostName));
#ifdef _DEBUG
   DBGPRINTF((OST, "-- gethostname = %.*s\n", sizeof(SMTPinit->HostName), SMTPinit->HostName));
#endif

	SMTPinit->A.sin_family   =  AF_INET;
   SMTPinit->A.sin_port     =  htons(SMTPhostPort);

   if (SMTPinit->A.sin_port == 0) {
		*Reason = TCP_BADPORT;
		free(SMTPinit);
#ifdef _DEBUG
      DBGPRINTF((OST, "<- SmtpInit : TCP_BADPORT\n"));
#endif
      return NULL;
	}

   SMTPinit->H = gethostbyname(SMTPhost);

   if (SMTPinit->H != NULL)
      memcpy (& SMTPinit->A.sin_addr.s_addr, SMTPinit->H->h_addr, SMTPinit->H->h_length);
   else {
#ifdef _DEBUG
      DBGPRINTF((OST, "<- SmtpInit : TCP_HOSTUNKNOWN\n"));
#endif
		*Reason = TCP_HOSTUNKNOWN;
		free(SMTPinit);
      return NULL;
   }

#ifdef _DEBUG
   DBGPRINTF((OST, "<-- SmtpInit\n"));
#endif

	*Reason = SMTP_SUCCESS;
	return SMTPinit;
}


int SmtpCleanUp(void)
{
	int rc = -1;

#ifdef _DEBUG
	DebugTime();               
   DBGPRINTF((OST, "--> SmtpCleanUp\n"));
#endif

	rc = TcpCleanUp();

#ifdef _DEBUG
	DebugTime();               
   DBGPRINTF((OST, "<-- SmtpCleanUp, rc = %d\n", rc));
#endif

	return rc;

}







